import logging
import random
from itertools import cycle
from typing import Any, Callable, Generator, Iterable, Sized


from qgis.PyQt.QtCore import QVariant


from qgis.PyQt.QtGui import QColor


from qgis.core import (
    QgsCategorizedSymbolRenderer,
    QgsExpression,
    QgsExpressionContext,
    QgsExpressionContextUtils,
    QgsFeature,
    QgsFeatureRequest,
    QgsProject,
    QgsRendererCategory,
    QgsSimpleFillSymbolLayer,
    QgsSymbol,
    QgsVectorLayer,
    QgsVectorLayerUtils,
)


from qgis.utils import iface
from warg import QuadNumber, TripleNumber, n_uint_mix_generator_builder

__all__ = [
    "categorise_layer",
    "random_color_alpha_generator",
    "random_color_generator",
    "random_rgba",
    "random_rgb",
    "styled_field_value_categorised",
]

_logger = logging.getLogger(__name__)


def random_rgb(mix: TripleNumber = (255, 255, 255)) -> TripleNumber:
    """

    :param mix: The upper limit of each element of the RGB Tuple
    :return: random RGB Tuple
    """
    red = random.randrange(0, mix[0])
    green = random.randrange(0, mix[1])
    blue = random.randrange(0, mix[2])
    return red, green, blue


def random_rgba(mix: QuadNumber = (1, 1, 1, 1)) -> QuadNumber:
    """

    :param mix: The upper limit of each element of the RGBA Tuple
    :return: random RGBA Tuple
    """
    red = random.randrange(0, mix[0])
    green = random.randrange(0, mix[1])
    blue = random.randrange(0, mix[2])
    alpha = random.randrange(0, mix[3])
    return red, green, blue, alpha


def random_color_generator(
    mix: TripleNumber = (1, 1, 1)
) -> Generator[TripleNumber, None, None]:
    """

    :param mix: The upper limit of each element of the RGBA Tuple
    :return: A generator of random RGB Tuples
    """
    while 1:
        yield random_rgb(mix=mix)


def random_color_alpha_generator(
    mix: QuadNumber = (1, 1, 1, 1)
) -> Generator[QuadNumber, None, None]:
    """

    :param mix: The upper limit of each element of the RGBA Tuple
    :return: A generator of random RGBA Tuples
    """
    while 1:
        yield random_rgba(mix=mix)


def categorise_layer(
    layer: QgsVectorLayer,
    field_name: str = "layer",
    *,
    color_iterable: Iterable = n_uint_mix_generator_builder(
        255, 255, 255, mix_min=(0, 0, 0)
    ),
    opacity: float = 1.0,
    outline_only: bool = False,
    outline_width=1.0,
) -> None:
    """

    https://qgis.org/pyqgis/3.0/core/Vector/QgsVectorLayer.html
    https://qgis.org/pyqgis/3.0/core/other/QgsFields.html

    :param outline_width:
    :param outline_only:
    :param opacity:
    :param layer:
    :param field_name:
    :param color_iterable:
    :return:
    """

    if isinstance(color_iterable, Sized):
        # noinspection PyTypeChecker
        color_iterable = cycle(color_iterable)

    if isinstance(color_iterable, Callable) and not isinstance(
        color_iterable, Generator
    ):
        # noinspection PyCallingNonCallable
        color_iterable = color_iterable()  # Compat

    color_iter = iter(color_iterable)

    available_field_names = layer.fields().names()

    assert (
        field_name in available_field_names
    ), f"Did not find {field_name=} in {available_field_names=}"

    render_categories = []
    for cat in layer.uniqueValues(layer.fields().indexFromName(field_name)):
        if cat is not None:
            sym = QgsSymbol.defaultSymbol(layer.geometryType())
            if sym is not None:
                set_symbol_styling(
                    color_iter, opacity, outline_only, outline_width, sym
                )

                render_categories.append(
                    QgsRendererCategory(cat, symbol=sym, label=str(cat), render=True)
                )

    if True:  # add default
        sym = QgsSymbol.defaultSymbol(layer.geometryType())
        # https://qgis.org/pyqgis/master/core/QgsSymbolLayer.html#qgis.core.QgsSymbolLayer.setFillColor
        # https://qgis.org/pyqgis/3.40/core/QgsSimpleLineSymbolLayer.html
        # QgsLinePatternFillSymbolLayer

        # sym.symbolLayer(0).setStrokeColor(QColor(*col))
        # StrokeWidth
        # StrokeStyle
        # FillColor
        # FillStyle

        if sym is not None:
            set_symbol_styling(color_iter, opacity, outline_only, outline_width, sym)

            render_categories.append(
                QgsRendererCategory(
                    QVariant(""), symbol=sym, label="default", render=True
                )
            )

        if False:
            # render_categories.append(QgsRendererCategory()) # crashes qgis
            render_categories.append(
                QgsRendererCategory([], symbol=sym, label="EmptyList", render=True)
            )
            render_categories.append(
                QgsRendererCategory("", symbol=sym, label="EmptyString", render=True)
            )
            render_categories.append(
                QgsRendererCategory("None", symbol=sym, label="None", render=True)
            )

    layer.setRenderer(QgsCategorizedSymbolRenderer(field_name, render_categories))
    layer.triggerRepaint()
    iface.layerTreeView().refreshLayerSymbology(layer.id())


def styled_field_value_categorised(
    layer: Any, style_attributes_layer, field_name="location_type"
) -> None:
    expression_str = f'represent_value("{field_name}")'
    # expression_str_unquoted = f'represent_value({field_name})'

    asd = QgsVectorLayerUtils.getValues(layer, expression_str, selectedOnly=False)
    cats = asd[0]

    if True:
        render_categories = []
        added_references = set()

        style_features = {
            f["admin_id"]: f for f in style_attributes_layer.getFeatures()
        }

        # refs = [r.lstrip('{').rstrip('}') for r in refs]
        refs = QgsVectorLayerUtils.getValues(layer, field_name, selectedOnly=False)[0]

        assert len(refs) == len(cats), f"{refs=}, {cats=}"

        for ref, cat in zip(refs, cats, strict=True):
            if False:
                _logger.error(cat)

            if cat is not None and ref not in added_references:
                sym = QgsSymbol.defaultSymbol(layer.geometryType())

                # Apply style from reference feature if available
                if ref in style_features:
                    style_feature = style_features[ref]
                    key = "display_rule.polygon.fillColor"
                    if key in style_feature.fields().names():
                        fill_color = str(style_feature[key])
                        if "#" in fill_color:
                            if False:
                                _logger.error(
                                    f"Set fill_color {fill_color} for {cat} in layer {layer.id()}"
                                )
                            sym.setColor(QColor(fill_color))
                        else:
                            if False:
                                _logger.error(
                                    f"Did not set fill_color for {cat} in layer {layer.id()}"
                                )

                render_categories.append(
                    QgsRendererCategory(
                        QVariant(ref), symbol=sym, label=str(cat), render=True, uuid=ref
                    )
                )
                added_references.add(ref)

        sym = QgsSymbol.defaultSymbol(layer.geometryType())
        render_categories.append(
            QgsRendererCategory(
                QVariant(None),
                symbol=sym,
                label="",
                render=True,
                # uuid=ref
            )
        )

        # renderer = QgsCategorizedSymbolRenderer(expression_str, render_categories) # DOES NOT WORK PROPER

        renderer = QgsCategorizedSymbolRenderer(field_name, render_categories)

    else:
        renderer = QgsCategorizedSymbolRenderer()
        renderer.setClassAttribute(expression_str)

        for c in renderer.createCategories(
            cats, QgsSymbol.defaultSymbol(layer.geometryType())
        ):
            renderer.addCategory(c)

    # renderer.filter()
    # renderer.setUsingSymbolLevels()

    # renderer.setOrderBy(QgsFeatureRequest.OrderBy([QgsFeatureRequest.OrderByClause(expression_str, False)]))
    # renderer.setOrderByEnabled(True)

    layer.setRenderer(renderer)
    layer.triggerRepaint()
    if False:
        iface.layerTreeView().refreshLayerSymbology(layer.id())
        iface.layerTreeView().layerTreeModel().refreshLayerLegend(
            QgsProject.instance().layerTreeRoot().findLayer(layer.id())
        )
        # iface.layerTreeView().layerTreeModel().recursivelyEmitDataChanged()


def set_symbol_styling(
    color_iter, opacity: float, outline_only: bool, outline_width: float, sym: Any
) -> None:
    col = next(color_iter)
    if len(col) == 3:
        col = (*col, 255)
    if outline_only:
        outline_symbol_layer = QgsSimpleFillSymbolLayer()
        outline_symbol_layer.setColor(QColor("transparent"))
        outline_symbol_layer.setStrokeWidth(outline_width)
        outline_symbol_layer.setStrokeColor(QColor(*col))
        sym.changeSymbolLayer(0, outline_symbol_layer)
    else:
        sym.setColor(QColor(*col))
    sym.setOpacity(opacity)



from qgis.PyQt.QtCore import QEvent


from qgis.PyQt.QtWidgets import QMessageBox, QTextEdit
from typing import Any

try:
    _layout_event = QEvent.Type
except:
    _layout_event = QEvent


__all__ = ["ResizableMessageBox"]


class ResizableMessageBox(QMessageBox):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setSizeGripEnabled(True)

    def event(self, event: Any) -> Any:
        if event.type() in (_layout_event.LayoutRequest, _layout_event.Resize):
            if event.type() == _layout_event.Resize:
                res = super().event(event)
            else:
                res = False

            details = self.findChild(QTextEdit)

            if details:
                details.setMaximumSize(16777215, 16777215)

            self.setMaximumSize(16777215, 16777215)
            return res

        return super().event(event)



from pathlib import Path


from qgis.PyQt.QtGui import QTransform
from typing import Any

__all__ = ["read_wld_file"]


def read_wld_file(geom: Any, wld_file_path: Path) -> None:
    assert wld_file_path is not None
    assert wld_file_path.exists()

    with open(wld_file_path) as wld_file:
        m32 = (float(c) for c in wld_file.readlines())

        transformer = QTransform(*m32)

        geom.transform(transformer)



import logging
import time
from itertools import tee
from typing import Any, Callable, Iterable, List, Mapping, Optional, Union
from warg import TripleNumber, n_uint_mix_generator_builder, passes_kws_to

from jord.typing_utilities.type_solving import (
    solve_attribute_uri,
    solve_field_uri,
    solve_qgis_type,
    solve_type_configuration,
    to_string_if_not_of_exact_type,
)

APPEND_TIMESTAMP = True
SKIP_MEMORY_LAYER_CHECK_AT_CLOSE = True
PIXEL_SIZE = 1
DEFAULT_NUMBER = 0
CONTRAST_ENHANCE = True
DEFAULT_LAYER_NAME = "TemporaryLayer"
DEFAULT_LAYER_CRS = "EPSG:4326"
VERBOSE = False
STRICT = False  # TODO: SET TO TRUE!
USE_TEMP_GROUP = False

__all__ = [
    "add_qgis_single_feature_layer",
    "add_qgis_single_geometry_layers",
    "add_qgis_multi_feature_layer",
]

_logger = logging.getLogger(__name__)


def add_qgis_single_feature_layer(
    qgis_instance_handle: Optional[Any],
    geom: Any,  #: QgsGeometry,
    name: Optional[str] = None,
    crs: Optional[str] = None,
    columns: Optional[Mapping[str, Any]] = None,
    index: bool = False,
    categorise_by_attribute: Optional[str] = None,
    color_generator: Callable[[], TripleNumber] = n_uint_mix_generator_builder(
        255, 255, 255, mix_min=(222, 222, 222)
    ),
    group: Any = None,
    visible: bool = True,
    opacity: float = 1.0,
    measurements: Optional[Iterable[Iterable[float]]] = None,
) -> List:
    """
    An example url is “Point?crs=epsg:4326&field=id:integer&field=name:string(20)&index=yes”

    :param opacity:
    :param color_generator:
    :param visible:
    :param categorise_by_attribute:
    :param group:
    :param columns: Field=name:type(length,precision) Defines an attribute of the layer. Multiple field
    parameters can be added to the data provider definition. Type is one of “integer”, “double”, “string”.
    :param index: index=yes Specifies that the layer will be constructed with a spatial index
    :param qgis_instance_handle:
    :param geom:
    :type geom: QgsGeometry
    :param name:
    :type name: Optional[str]
    :param crs: Crs=definition Defines the coordinate reference system to use for the layer. Definition is any
    string accepted by QgsCoordinateReferenceSystem.createFromString()
    :param measurements: UNUSED!
    :return: None
    :rtype: None
    """

    from qgis.core import (
        QgsVectorLayer,
        QgsFeature,
        QgsVectorLayer,
        QgsRasterLayer,
        QgsProject,
        QgsWkbTypes,
    )
    from jord.qgis_utilities import categorise_layer

    import qgis

    # uri = geom.type()
    # uri = geom.wkbType()
    # uri = geom.wktTypeStr()

    return_collection = []

    if geom.wkbType() == QgsWkbTypes.NoGeometry:
        return return_collection

    geom_type = QgsWkbTypes.displayString(geom.wkbType())

    if True:
        _logger.error(f"{geom},{geom_type=}")

    uri = geom_type  # TODO: URI MIGHT BE NONE?

    if name is None:
        name = DEFAULT_LAYER_NAME

    if crs is None:
        crs = DEFAULT_LAYER_CRS

    layer_name = f"{name}"
    if APPEND_TIMESTAMP:
        layer_name += f"_{time.time()}"

    if columns:
        fields = {k: solve_qgis_type(v) for k, v in columns}
        field_type_configuration = {
            k: solve_type_configuration(v, k, columns) for k, v in columns.items()
        }
    else:
        fields = None
        field_type_configuration = None

    if categorise_by_attribute and fields:
        assert (
            categorise_by_attribute in fields
        ), f"{categorise_by_attribute} was not found in {fields}"

    if qgis_instance_handle is None:
        qgis_project = QgsProject.instance()
    elif not isinstance(qgis_instance_handle, QgsProject):
        qgis_project = qgis_instance_handle.qgis_project
    else:
        qgis_project = qgis_instance_handle

    if geom_type in (
        QgsWkbTypes.GeometryCollection,
        QgsWkbTypes.GeometryCollectionZ,
        QgsWkbTypes.GeometryCollectionM,
        QgsWkbTypes.GeometryCollectionZM,
    ):
        for g in geom.asGeometryCollection():  # TODO: Look into recursion?
            uri = QgsWkbTypes.displayString(g.wkbType())

            if True:
                _logger.error(f"{g},{uri=}")

            sub_type = uri  # TODO: URI MIGHT BE NONE?

            uri += "?"

            if crs:
                uri += f"crs={crs}"

            if fields:
                uri = str(uri).rstrip("&")
                for k, v in fields.items():
                    uri += f"&field={k}:{v}"

            uri = str(uri).rstrip("&")
            uri += f'&index={"yes" if index else "no"}'
            uri.replace("?&", "?")

            feat = QgsFeature()
            feat.setGeometry(g)

            if columns:
                feat.initAttributes(len(columns))

                if STRICT:
                    for field_idx, attr in enumerate(columns.values()):
                        feat.setAttribute(field_idx, attr)
                else:
                    feat.setAttributes(
                        list(to_string_if_not_of_exact_type(columns.values()))
                    )

            sub_layer = QgsVectorLayer(uri, f"{layer_name}_{sub_type}", "memory")
            layer_data_provider = sub_layer.dataProvider()
            layer_data_provider.addFeatures([feat])
            layer_data_provider.updateExtents()

            if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
                sub_layer.setCustomProperty("skipMemoryLayersCheck", 1)

            if categorise_by_attribute:
                categorise_layer(
                    sub_layer,
                    categorise_by_attribute,
                    color_iterable=color_generator,
                    opacity=opacity,
                )

            sub_layer.commitChanges()
            sub_layer.updateFields()
            sub_layer.updateExtents()

            return_collection.append(sub_layer)

            if group:
                qgis_project.addMapLayer(sub_layer, False)
                group.insertLayer(0, sub_layer)
            else:
                qgis_project.addMapLayer(sub_layer)

            layer_tree_handle = qgis_project.layerTreeRoot().findLayer(sub_layer.id())
            if layer_tree_handle:
                layer_tree_handle.setItemVisibilityChecked(visible)
    else:
        uri += "?"

        if crs:
            uri += f"crs={crs}"

        if fields:
            uri = solve_field_uri(field_type_configuration, fields, uri)

        uri = str(uri).rstrip("&")
        uri += f'&index={"yes" if index else "no"}'
        uri.replace("?&", "?")

        feat = QgsFeature()
        feat.setGeometry(geom)

        if columns:
            feat.initAttributes(len(columns))

            if False:
                for field_idx, attr in enumerate(columns.values()):
                    feat.setAttribute(field_idx, attr)
            else:
                feat.setAttributes(
                    list(to_string_if_not_of_exact_type(columns.values()))
                )

        layer = QgsVectorLayer(uri, layer_name, "memory")
        layer_data_provider = (
            layer.dataProvider()
        )  # DEFAULT DATA PROVIDER, TODO: MAYBE CHANGE THIS
        assert feat.isValid(), f"{feat} was invalid"
        res, out_feats = layer_data_provider.addFeatures([feat])

        if not res:
            _logger.error(f"{layer_data_provider.lastError()}")

            assert (
                res
            ), f"Failure while adding features {res} {layer_data_provider.lastError()}"

        layer_data_provider.updateExtents()

        if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
            layer.setCustomProperty("skipMemoryLayersCheck", 1)

        if categorise_by_attribute:
            categorise_layer(
                layer,
                categorise_by_attribute,
                color_iterable=color_generator,
                opacity=opacity,
            )

        layer.commitChanges()
        layer.updateFields()
        layer.updateExtents()

        return_collection.append(layer)

        if group:
            qgis_project.addMapLayer(layer, False)
            group.insertLayer(0, layer)
        else:
            qgis_project.addMapLayer(layer)

        layer_tree_handle = qgis_project.layerTreeRoot().findLayer(layer.id())
        if layer_tree_handle:
            layer_tree_handle.setItemVisibilityChecked(visible)

    actions = qgis.utils.iface.layerTreeView().defaultActions()
    actions.showFeatureCount()
    actions.showFeatureCount()  # TODO: Duplicate?

    return return_collection


@passes_kws_to(add_qgis_single_feature_layer)
def add_qgis_single_geometry_layers(
    qgis_instance_handle: Any, geoms: Mapping, **kwargs  # [str,QgsGeometry]
) -> None:
    """

    :param qgis_instance_handle:
    :param geoms:
    :param kwargs:
    :return:
    """
    for name, geom in geoms.items():
        add_qgis_single_feature_layer(qgis_instance_handle, geom, name, **kwargs)


def add_qgis_multi_feature_layer(
    qgis_instance_handle: Any,
    geoms: Iterable,  # [QgsGeometry]
    name: Optional[Iterable[str]] = None,
    crs: Optional[str] = None,
    columns: Optional[
        Union[Mapping[str, Mapping[str, Any]], Iterable[Mapping[str, Any]]]
    ] = None,
    categorise_by_attribute: Optional[str] = None,
    color_generator: Iterable = n_uint_mix_generator_builder(
        255, 255, 255, mix_min=(255 // 2, 255 // 2, 255 // 2)
    ),
    index: bool = False,
    group: Any = None,
    visible: bool = True,
    opacity: float = 1.0,
    measurements: Optional[Iterable[Iterable[float]]] = None,
) -> Optional[List]:
    """

        fields  == column definition name, type, length/size
        Multiple field parameters can be added to the data provider definition. type is one of “integer”,
        “double”, “string”.

    An example url is “Point?crs=epsg:4326&field=id:integer&field=name:string(20)&index=yes”

    :param measurements: UNUSED!
    :param opacity:
    :param color_generator:
    :param categorise_by_attribute:
    :param visible:
    :param group:
    :param qgis_instance_handle:
    :param geoms:
    :param name:
    :param crs:
    :param columns:
    :param index:
    :return:
    """

    from jord.qgis_utilities.categorisation import categorise_layer

    from qgis.core import (
        QgsFeature,
        QgsVectorLayer,
        QgsProject,
        QgsFeatureSink,
        QgsWkbTypes,
    )

    import qgis

    # uri = geom.type()
    # uri = geom.wkbType()
    # uri = geom.wktTypeStr()

    return_collection = []

    if name is None:
        name = DEFAULT_LAYER_NAME

    if crs is None:
        crs = DEFAULT_LAYER_CRS

    layer_name = f"{name}"
    if APPEND_TIMESTAMP:
        layer_name += f"_{time.time()}"

    geom_type = None
    uri = None

    num_cols = None
    attr_generator = None
    fields = None
    field_type_configuration = None

    if columns and len(columns):  # TODO: FIND MAX LENGTH STR
        if isinstance(columns, Mapping):
            attr_generator, attr_type_sampler = tee(iter(columns.values()))
        elif isinstance(columns, Iterable):
            attr_generator, attr_type_sampler = tee(iter(columns))
        else:
            raise TypeError(f"columns must be a mapping or an iterable of mappings")

        field_type_configuration, fields, num_cols = solve_attribute_uri(
            attr_type_sampler, columns
        )

    if categorise_by_attribute and fields:
        assert (
            categorise_by_attribute in fields
        ), f"{categorise_by_attribute} was not found in {fields}"

    if not geoms:
        # logger.info(f"Found no geometries, {geoms} for {name}")
        return  # No geometry

    features = []
    if not isinstance(geoms, Iterable):
        geoms = [geoms]

    for geom in geoms:
        # geom:QgsGeometry

        if geom.wkbType() == QgsWkbTypes.NoGeometry:
            continue

        geom_type_ = QgsWkbTypes.displayString(geom.wkbType())

        if geom_type is None:
            geom_type = geom_type_

        if uri is None:
            uri = geom_type_  # TODO: URI MIGHT BE NONE?

        assert (
            geom_type == geom_type_
        ), f"{geom_type_} is the not the same geometry type as {geom_type}"

        if geom_type in (  # TODO: VERIFY LOGIC!
            QgsWkbTypes.GeometryCollection,
            QgsWkbTypes.GeometryCollectionZ,
            QgsWkbTypes.GeometryCollectionM,
            QgsWkbTypes.GeometryCollectionZM,
        ):
            for g in geom.asGeometryCollection():  # TODO: Look into recursion?

                sub_type = QgsWkbTypes.displayString(g.wkbType())

                return_collection.extend(
                    add_qgis_multi_feature_layer(
                        qgis_instance_handle,
                        g,
                        f"{name}_{sub_type}",
                        crs=crs,
                        columns=columns,
                        categorise_by_attribute=categorise_by_attribute,
                        color_generator=color_generator,
                        index=index,
                        group=group,
                        visible=visible,
                        opacity=opacity,
                        measurements=measurements,  # TODO: THIS WILL BE WEIRD!
                    )
                )
            return return_collection
        else:
            feat = QgsFeature()

            if attr_generator:
                row = next(attr_generator, None)
                if row:
                    feat.initAttributes(num_cols)
                    feat.setAttributes(
                        list(to_string_if_not_of_exact_type(row.values()))
                    )

            feat.setGeometry(geom)

            assert feat.isValid(), f"{feat} was invalid"
            features.append(feat)

    if STRICT:
        assert len(list(geoms)) == len(
            features
        ), f"Some features where dropped! {len(list(geoms))} != {len(features)}"
    else:
        if len(list(geoms)) != len(features):
            _logger.error(
                f"Some features where dropped! {len(list(geoms))} != {len(features)}"
            )

    if uri is None:
        if STRICT:
            raise Exception("uri is None")
        else:
            return return_collection

    uri += "?"

    if crs:
        uri += f"crs={crs}"

    if fields:  # field=name:type(length,precision)
        uri = solve_field_uri(field_type_configuration, fields, uri)

    uri = str(uri).rstrip("&")
    uri += f'&index={"yes" if index else "no"}'
    uri.replace("?&", "?")

    layer = QgsVectorLayer(uri, layer_name, "memory")
    layer_data_provider = layer.dataProvider()
    # pr.addAttributes([QgsField("name", QVariant.String),QgsField("age", QVariant.Int),QgsField("size",
    # QVariant.Double)])

    (res, out_feats) = layer_data_provider.addFeatures(
        features
        # , QgsFeatureSink.RollBackOnErrors
    )

    if not res:
        _logger.error(f"{layer_data_provider.lastError()}")

        if not res:
            msg = (
                f"Failure while adding features {res} {layer_data_provider.lastError()}"
            )
            assert res, msg
            _logger.warning(msg)

        assert len(list(geoms)) == len(
            out_feats
        ), f"Some features where dropped! return status {res}:  {len(list(geoms))} != {len(out_feats)}"

    if len(list(geoms)) != layer.featureCount():
        _logger.error(f"{features}")

    if STRICT:
        assert (
            len(list(geoms)) == layer.featureCount()
        ), f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
    else:
        if len(list(geoms)) != layer.featureCount():
            _logger.error(
                f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
            )

    layer_data_provider.updateExtents()

    if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
        layer.setCustomProperty("skipMemoryLayersCheck", 1)

    if categorise_by_attribute:
        categorise_layer(
            layer,
            categorise_by_attribute,
            color_iterable=color_generator,
            opacity=opacity,
        )

    if STRICT:
        assert (
            len(list(geoms)) == layer.featureCount()
        ), f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
    else:
        if len(list(geoms)) != layer.featureCount():
            _logger.error(
                f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
            )

    layer.commitChanges()
    layer.updateFields()
    layer.updateExtents()

    if STRICT:
        assert (
            len(list(geoms)) == layer.featureCount()
        ), f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
    else:
        if len(list(geoms)) != layer.featureCount():
            _logger.error(
                f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
            )

    if qgis_instance_handle is None:
        qgis_project = QgsProject.instance()
    elif not isinstance(qgis_instance_handle, QgsProject):
        qgis_project = qgis_instance_handle.qgis_project
    else:
        qgis_project = qgis_instance_handle

    if group:
        qgis_project.addMapLayer(layer, False)
        group.insertLayer(0, layer)
    else:
        qgis_project.addMapLayer(layer)

    layer_tree_handle = qgis_project.layerTreeRoot().findLayer(layer.id())
    if layer_tree_handle:
        layer_tree_handle.setItemVisibilityChecked(visible)

    actions = qgis.utils.iface.layerTreeView().defaultActions()
    actions.showFeatureCount()  # TODO: Twice?
    actions.showFeatureCount()

    return_collection.append(layer)

    assert len(return_collection) > 0, f"Return collection was empty"

    return return_collection



from typing import Any, Collection

__all__ = ["set_column_visibility", "set_visible_columns"]

from jord.qgis_utilities.configuration import read_plugin_setting, store_plugin_setting


def set_column_visibility(layers: Any, column_name: str, visible: bool = False) -> None:
    if not isinstance(layers, Collection):
        layers = [layers]

    for layer in layers:
        config = layer.attributeTableConfig()
        columns = config.columns()
        for column in columns:
            if column.name == column_name:
                column.hidden = not visible
                break

        config.setColumns(columns)
        layer.setAttributeTableConfig(config)


def set_visible_columns(layers: Any, column_names: Collection[str]) -> None:
    if not isinstance(layers, Collection):
        layers = [layers]

    for layer in layers:
        config = layer.attributeTableConfig()
        columns = config.columns()
        for column in columns:
            if column.name in column_names:
                column.hidden = True
            else:
                column.hidden = False

        config.setColumns(columns)
        layer.setAttributeTableConfig(config)


def update_column_visibility_with_defaults(
    layers: Any,
    column_names: Collection[str],
    default_visible: bool = True,
    persist_settings: bool = True,
) -> None:
    """Update column visibility with defaults for new columns.

    Args:
        layers: Single layer or collection of layers
        column_names: Names of columns to manage visibility
        default_visible: Default visibility for new columns
        persist_settings: Whether to save settings
    """
    if not isinstance(layers, Collection):
        layers = [layers]

    for layer in layers:
        config = layer.attributeTableConfig()
        columns = config.columns()
        settings_key = f"attribute_table/{layer.id()}/column_visibility"

        # Load or initialize stored visibility settings
        stored_visibility = read_plugin_setting(settings_key, default_value={})

        for column in columns:
            column_name = column.name

            if column_name in stored_visibility:
                # Use stored setting
                column.hidden = not stored_visibility[column_name]
            else:
                # Apply default for new columns
                is_visible = (
                    column_name in column_names
                    if default_visible
                    else column_name not in column_names
                )
                column.hidden = not is_visible
                stored_visibility[column_name] = is_visible

        # Save updated settings
        if persist_settings:
            store_plugin_setting(settings_key, stored_visibility)

        config.setColumns(columns)
        layer.setAttributeTableConfig(config)



import logging
from enum import Enum, IntEnum


from qgis.core import (
    QgsDefaultValue,
    QgsDefaultValue,
    QgsEditorWidgetSetup,
    QgsEditorWidgetSetup,
    QgsFieldConstraints,
    QgsFieldConstraints,
    QgsMapLayer,
)
from typing import Any, Iterable, Mapping, Sequence, Type

from jord.qgis_utilities.helpers.widgets import (
    CHECKBOX_WIDGET,
    NULLABLE_CHECKBOX_WIDGET,
    UNIQUE_VALUES_WIDGET,
)

__all__ = [
    "set_field_widget",
    "make_field_unique",
    "make_field_not_null",
    "make_field_default",
    "make_field_boolean",
    "make_field_reuse_last_entered_value",
    "make_value_relation_widget",
    "make_enum_dropdown_widget",
    "make_iterable_dropdown_widget",
    "make_sorted_mapping_dropdown_widget",
    "make_value_map_widget",
    "make_field_datetime",
    "make_field_readonly",
    "make_external_resource_widget",
]


class DocumentViewerEnum(IntEnum):
    no_content = 0
    image = 1
    audio = 2
    video = 3
    web_view = 4


AUTO_DIMENSION = 0

ABSOLUTE_PATH = 0
RELATIVE_PATH = 1

FILE_PATHS = 0
DIRECTORY_PATHS = 1

IGNORE_THIS_STRING = """

Line edit – a simple edit box
Classification – displays a combo box with the values used for “unique value” classification (symbology tab)
Range – allows numeric values within a given range, the widget can be either slider or spin box
Unique values
    editable – displays a line edit widget with auto-completion suggesting values already used in the
    attribute table
    not editable – displays a combo box with already used values
File name – adds a file chooser dialog
Value map – shows a combo box with predefined description/value items
Enumeration – a combo box with values that can be used within the columns type
Immutable – read-only
Hidden – makes the attribute invisible for the user
CheckBox – a checkbox with customizable representation for both checked and unchecked state
Text edit – an edit box that allow multiple input lines
Calendar – a calendar widget to input dates
"""

IGNORE_THIS_STRING2 = """
QGIS Widget Types
_________________

Binary

Checkbox

Classification

Color

DateTime

Enumeration

Attachment

Geometry

Hidden

JsonView

KeyValue

List

Range

RelationReference

TextEdit

UniqueValues

UuidGenerator

ValueMap

ValueRelation

"""

_logger = logging.getLogger(__name__)


def set_field_widget(layers: Any, field_name: str, form_widget: Any) -> None:
    """
    https://gis.stackexchange.com/questions/470963/setting-dropdown-on-feature-attribute-form-using-plugin


    :param layers:
    :param field_name:
    :param form_widget:
    :return:
    """

    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        idx = layer.fields().indexFromName(field_name)
                        if idx < 0:
                            continue
                        layer.setEditorWidgetSetup(
                            idx,
                            form_widget,
                        )
            else:
                idx = layers_inner.fields().indexFromName(field_name)
                if idx < 0:
                    continue
                layers_inner.setEditorWidgetSetup(
                    idx,
                    form_widget,
                )


def make_field_unique(
    layers: Sequence[Any], *, field_name: str = "id", auto_generate: bool = True
) -> None:
    """

    :param layers:
    :param field_name:
    :param auto_generate:
    :return:
    """
    if layers is None:
        return

    unique_widget = None
    default_value_generator = None

    if False:
        unique_widget = QgsEditorWidgetSetup(
            "UuidGenerator",
            {},
        )
        if False:
            _logger.error(unique_widget.config())
    elif auto_generate:
        default_value_generator = QgsDefaultValue()
        default_value_generator.setExpression("rtrim( ltrim( uuid(), '{'), '}')")
    else:
        unique_widget = UNIQUE_VALUES_WIDGET

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        idx = layer.fields().indexFromName(field_name)
                        if idx < 0:
                            continue

                        if unique_widget:
                            layer.setEditorWidgetSetup(
                                idx,
                                unique_widget,
                            )
                        elif default_value_generator:
                            layer.setDefaultValueDefinition(
                                idx, default_value_generator
                            )
                        else:
                            raise NotImplementedError

                        layer.setFieldConstraint(
                            idx, QgsFieldConstraints.ConstraintNotNull
                        )
                        layer.setFieldConstraint(
                            idx, QgsFieldConstraints.ConstraintUnique
                        )
            else:
                idx = layers_inner.fields().indexFromName(field_name)

                if idx < 0:
                    continue

                if unique_widget:
                    layers_inner.setEditorWidgetSetup(
                        idx,
                        unique_widget,
                    )
                elif default_value_generator:
                    layers_inner.setDefaultValueDefinition(idx, default_value_generator)
                else:
                    raise NotImplementedError

                layers_inner.setFieldConstraint(
                    idx, QgsFieldConstraints.ConstraintNotNull
                )
                layers_inner.setFieldConstraint(
                    idx, QgsFieldConstraints.ConstraintUnique
                )


def make_field_not_null(layers: Sequence[Any], field_name: str = "name") -> None:
    """

    :param layers:
    :param field_name:
    :return:
    """
    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layers in layers_inner:
                    if layers:
                        idx = layers.fields().indexFromName(field_name)
                        if idx < 0:
                            continue

                        layers.setFieldConstraint(
                            idx, QgsFieldConstraints.ConstraintNotNull
                        )
            else:
                idx = layers_inner.fields().indexFromName(field_name)
                if idx < 0:
                    continue
                layers_inner.setFieldConstraint(
                    idx, QgsFieldConstraints.ConstraintNotNull
                )


def make_field_default(
    layers: Sequence[Any], field_name: str, default_expression: str = "'None'"
) -> None:
    """

    :param layers:
    :param field_name:
    :param default_expression:
    :return:
    """
    if layers is None:
        return

    default_value = QgsDefaultValue()
    default_value.setExpression(default_expression)

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layers in layers_inner:
                    if layers:
                        layers.setDefaultValueDefinition(
                            layers.fields().indexFromName(field_name), default_value
                        )
            else:
                layers_inner.setDefaultValueDefinition(
                    layers_inner.fields().indexFromName(field_name), default_value
                )


def make_field_boolean(
    layers: Sequence[Any], field_name: str, nullable: bool = True
) -> None:
    """

    :param nullable:
    :param layers:
    :param field_name:
    :return:
    """
    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        idx = layer.fields().indexFromName(field_name)

                        if idx < 0:
                            continue

                        layer.setEditorWidgetSetup(
                            idx,
                            NULLABLE_CHECKBOX_WIDGET if nullable else CHECKBOX_WIDGET,
                        )
            else:
                idx = layers_inner.fields().indexFromName(field_name)

                if idx < 0:
                    continue

                layers_inner.setEditorWidgetSetup(
                    idx,
                    NULLABLE_CHECKBOX_WIDGET if nullable else CHECKBOX_WIDGET,
                )


def make_field_reuse_last_entered_value(layers: Sequence[Any], field_name: str) -> None:
    """

    :param layers:
    :param field_name:
    :return:
    """
    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        idx = layer.fields().indexFromName(field_name)

                        if idx < 0:
                            continue

                        layer_form_config = layer.editFormConfig()
                        layer_form_config.setReuseLastValue(idx, True)
                        layer.setEditFormConfig(layer_form_config)
            else:
                idx = layers_inner.fields().indexFromName(field_name)

                if idx < 0:
                    continue

                layer_form_config = layers_inner.editFormConfig()
                layer_form_config.setReuseLastValue(idx, True)
                layers_inner.setEditFormConfig(layer_form_config)


def fit_field_to_length(layers: Sequence[Any], field_name: str, length: int) -> None:
    """

    :param layers:
    :param field_name:
    :param length:
    :return:
    """
    if layers is None:
        return

    for layers_inner in layers:
        if layers_inner:
            if isinstance(layers_inner, Iterable):
                for layer in layers_inner:
                    if layer:
                        fields = layer.fields()

                        idx = fields.indexFromName(field_name)
                        if idx < 0:
                            continue

                        fields[idx].setLength(length)
            else:
                fields = layers_inner.fields()
                idx = fields.indexFromName(field_name)
                if idx < 0:
                    continue
                fields[idx].setLength(length)


def make_value_map_widget(mapp: Mapping[str, Any]) -> Any:
    """

    :param mapp:
    :return:
    """
    return QgsEditorWidgetSetup(
        "ValueMap",
        {"map": mapp},
    )


def make_enum_dropdown_widget(_enum: Type[Enum]) -> Any:
    """

    :param _enum:
    :return:
    """
    return make_value_map_widget(
        {
            name: _enum.__getitem__(name).value
            for name in sorted({l.name for l in _enum})
        }
    )


def make_sorted_mapping_dropdown_widget(m: Mapping[str, Any]) -> Any:
    """

    :param m:
    :return:
    """
    return make_value_map_widget({k: m[k] for k in sorted(m)})


def make_iterable_dropdown_widget(it: Iterable) -> Any:
    """

    :param it:
    :return:
    """
    return make_value_map_widget({name: name for name in sorted({l for l in it})})


def make_value_relation_widget(
    target_layer_id: str,
    *,
    target_key_field_name: str = "key",
    target_value_field_name: str = "name",
    use_completer: bool = False,
    order_by_value: bool = True,
    allow_null_values: bool = False,
    allow_multiple_values: bool = False
) -> Any:
    """
      <editWidget type="ValueRelation">
            <config>
              <Option type="Map">
                <Option value="false" type="bool" name="AllowMulti"/>
                <Option value="false" type="bool" name="AllowNull"/>
                <Option value="2" type="int" name="CompleterMatchFlags"/>
                <Option value="&quot;name&quot;" type="QString" name="Description"/>
                <Option value="true" type="bool" name="DisplayGroupName"/>
                <Option value="" type="QString" name="FilterExpression"/>
                <Option value="admin_id" type="QString" name="Group"/>
                <Option value="admin_id" type="QString" name="Key"/>
                <Option value="location_types_1743074272_2672093_6cf6edd7_a194_42b5_a1a9_8ccec987325c"
                type="QString" name="Layer"/>
                <Option value="location_types_1743074272.2672093" type="QString" name="LayerName"/>
                <Option value="memory" type="QString" name="LayerProviderName"/>
                <Option value="None?crs=EPSG:4326&amp;field=admin_id:text(255)&amp;field=name:text(
                255)&amp;index=no&amp;uid={57bd270e-08b6-4e66-8a4f-81f703d3a3a5}" type="QString"
                name="LayerSource"/>
                <Option value="1" type="int" name="NofColumns"/>
                <Option value="true" type="bool" name="OrderByValue"/>
                <Option value="true" type="bool" name="UseCompleter"/>
                <Option value="name" type="QString" name="Value"/>
              </Option>
            </config>
          </editWidget>


    :param target_layer_id:
    :param target_key_field_name:
    :param target_value_field_name:
    :param use_completer:
    :param order_by_value:
    :param allow_null_values:
    :param allow_multiple_values:
    :return:
    """

    return QgsEditorWidgetSetup(
        "ValueRelation",
        {
            "AllowMulti": allow_multiple_values,
            "AllowNull": allow_null_values,
            "FilterExpression": "",
            "Key": target_key_field_name,
            "Layer": target_layer_id,
            "NofColumns": 1,
            "OrderByValue": order_by_value,
            "UseCompleter": use_completer,
            "Value": target_value_field_name,
        },
    )


def make_field_datetime(layer: Any, field_name: str) -> None:
    """

    :param layer:
    :param field_name:
    :return:
    """
    config = {
        "allow_null": True,
        "calendar_popup": True,
        "display_format": "yyyy-MM-dd HH:mm:ss",
        "field_format": "yyyy-MM-dd HH:mm:ss",
        "field_iso_format": False,
    }
    type = "DateTime"
    fields = layer.fields()
    field_idx = fields.indexOf(field_name)

    widget_setup = QgsEditorWidgetSetup(type, config)
    layer.setEditorWidgetSetup(field_idx, widget_setup)
    layer.setDefaultValueDefinition(field_idx, QgsDefaultValue("now()"))


def make_field_readonly(layer: Any, field_name: str, option: bool = True) -> None:
    """

    :param layer:
    :param field_name:
    :param option:
    :return:
    """
    if layer.type() != QgsMapLayer.VectorLayer:
        return

    fields = layer.fields()
    field_idx = fields.indexOf(field_name)
    if field_idx >= 0:
        form_config = layer.editFormConfig()
        form_config.setReadOnly(field_idx, option)
        layer.setEditFormConfig(form_config)


def make_external_resource_widget(
    document_viewer: DocumentViewerEnum = DocumentViewerEnum.image,
    document_viewer_width: int = AUTO_DIMENSION,
    document_viewer_height: int = AUTO_DIMENSION,
) -> Any:
    """


    <editWidget type="ExternalResource">
          <config>
            <Option type="Map">
              <Option name="DocumentViewer" type="int" value="0"/>
              <Option name="DocumentViewerHeight" type="int" value="0"/>
              <Option name="DocumentViewerWidth" type="int" value="0"/>
              <Option name="FileWidget" type="bool" value="true"/>
              <Option name="FileWidgetButton" type="bool" value="true"/>
              <Option name="FileWidgetFilter" type="QString" value=""/>
              <Option name="PropertyCollection" type="Map">
                <Option name="name" type="QString" value=""/>
                <Option name="properties"/>
                <Option name="type" type="QString" value="collection"/>
              </Option>
              <Option name="RelativeStorage" type="int" value="0"/>
              <Option name="StorageAuthConfigId" type="QString" value=""/>
              <Option name="StorageMode" type="int" value="0"/>
              <Option name="StorageType" type="QString" value=""/>
            </Option>
          </config>
        </editWidget>

    :param document_viewer:
    :param document_viewer_width:
    :param document_viewer_height:
    :return:
    """

    return QgsEditorWidgetSetup(
        "ExternalResource",
        {
            "FileWidget": True,
            "FileWidgetButton": True,
            "FileWidgetFilter": "",
            "DocumentViewer": document_viewer.value,
            "DocumentViewerHeight": document_viewer_height,
            "DocumentViewerWidth": document_viewer_width,
            "RelativeStorage": ABSOLUTE_PATH,
            "StorageMode": FILE_PATHS,
        },
    )


IGNORE_THIS_STRING3 = """

ews = layer.editorWidgetSetup(field_index)
print("Type:", ews.type())
print("Config:", ews.config())
"""



import logging


from qgis.PyQt.QtGui import QColor


from qgis.core import (
    QgsCategorizedSymbolRenderer,
    QgsExpression,
    QgsExpressionContext,
    QgsExpressionContextUtils,
    QgsFeature,
    QgsFeatureRequest,
    QgsProject,
    QgsRendererCategory,
    QgsSimpleFillSymbolLayer,
    QgsSymbol,
    QgsVectorLayer,
    QgsVectorLayerUtils,
)


from qgis.utils import iface

_logger = logging.getLogger(__name__)


def styled_field_value_categorised2(
    layer, style_attributes_layer, field_name="location_type"
):
    expression_str = f'represent_value("{field_name}")'

    render_categories = []
    added_categories = set()

    style_features = {f["admin_id"]: f for f in style_attributes_layer.getFeatures()}

    for ref, cat in zip(
        QgsVectorLayerUtils.getValues(layer, field_name, selectedOnly=False)[0],
        QgsVectorLayerUtils.getValues(layer, expression_str, selectedOnly=False)[0],
        strict=True,
    ):
        _logger.error(cat)

        if cat is not None and cat not in added_categories:
            sym = QgsSymbol.defaultSymbol(layer.geometryType())

            # Apply style from reference feature if available
            if ref in style_features:
                style_feature = style_features[ref]
                key = "display_rule.polygon.fillColor"
                if key in style_feature.fields().names():
                    fill_color = str(style_feature[key])
                    if "#" in fill_color:
                        _logger.error(
                            f"Set fill_color {fill_color} for {cat} in layer {layer.id()}"
                        )
                        sym.setColor(QColor(fill_color))
                    else:
                        _logger.error(
                            f"Did not set fill_color for {cat} in layer {layer.id()}"
                        )

            render_categories.append(
                QgsRendererCategory(ref, symbol=sym, label=cat, render=True)
            )
            added_categories.add(cat)

    sym = QgsSymbol.defaultSymbol(layer.geometryType())
    render_categories.append(
        QgsRendererCategory("", symbol=sym, label="default", render=True)
    )

    rendered = QgsCategorizedSymbolRenderer(field_name, render_categories)

    # rendered.setClassAttribute(expression_str)
    # rendered.rebuildHash()

    # rendered.matchToSymbols()

    layer.setRenderer(rendered)
    layer.triggerRepaint()
    iface.layerTreeView().refreshLayerSymbology(layer.id())

    if False:  # TODO: DOES NOT WORK
        # Add this to update category counts
        layer_tree_layer = (
            iface.layerTreeView().layerTreeModel().rootGroup().findLayer(layer)
        )
        if layer_tree_layer:
            layer_tree_layer.setCustomProperty("showFeatureCount", True)
            iface.layerTreeView().layerTreeModel().refreshLayerLegend(layer_tree_layer)


def expression_field_value_categorised_do_not_use(layer, field_name="location_type"):
    expression_str = f'represent_value("{field_name}")'

    # expression = QgsExpression(expression_str)
    # context = QgsExpressionContext()

    # context.appendScope(QgsExpressionContextUtils.globalScope())
    # context.appendScope(QgsExpressionContextUtils.projectScope(QgsProject.instance()))
    # context.appendScope(QgsExpressionContextUtils.atlasScope(None))

    # context.appendScope(QgsExpressionContextUtils.layerScope(layer))
    # context.appendScope(QgsExpressionContextUtils.layerScope(reference_layer))

    render_categories = []
    added_categories = set()

    # for feature in layer.getFeatures(QgsFeatureRequest(expression)):
    # feature_iterator = QgsVectorLayerUtils.getValuesIterator(layer, expression_str, selectedOnly=False)
    # while feature_iterator.isValid():
    # cat = QgsFeature()
    # feature_iterator.nextFeature(cat)

    for cat in QgsVectorLayerUtils.getValues(layer, expression_str, selectedOnly=False)[
        0
    ]:
        _logger.error(cat)

        if cat is not None and cat not in added_categories:
            sym = QgsSymbol.defaultSymbol(layer.geometryType())

            # context.setFeature(feature)

            # context.setFeature(QgsFeature())
            # context.lastScope().setVariable(field_name, cat)

            # label = expression.evaluate(context)

            # if expression.hasEvalError():
            #  logger.error(f"Expression evaluation error: {expression.evalErrorString()} {label}")

            render_categories.append(
                QgsRendererCategory(cat, symbol=sym, label=cat, render=True)
            )
            added_categories.add(cat)

    rendered = QgsCategorizedSymbolRenderer(expression_str, render_categories)
    layer.setRenderer(rendered)
    layer.triggerRepaint()
    iface.layerTreeView().refreshLayerSymbology(layer.id())



import pandas
from pandas import DataFrame
from typing import Any, Collection, List, Mapping, Optional

__all__ = ["df_to_columns", "columns_to_df", "normalize_na"]


def df_to_columns(
    shape_df: DataFrame, ignored_columns: Optional[Collection] = None
) -> List[Mapping[str, Any]]:
    columns = []

    for key, c in shape_df.iterrows():
        o = {"key": key, **dict(c)}

        if ignored_columns:
            for p in ignored_columns:
                o.pop(p)

        columns.append(o)

    return columns


def columns_to_df(columns: List[Mapping[str, Any]]) -> DataFrame: ...


def normalize_na(d: Any) -> Optional[str]:
    if pandas.isna(d):
        return None

    # If needed, implement proper display rule normalization logic here
    # This could involve parsing the string representation and ensuring
    # extrusion is either None or properly initialized

    return str(d)



import logging
import shapely


from qgis.PyQt.QtCore import QDateTime, QVariant
from typing import Any, Generator, Mapping, Optional, Tuple

_logger = logging.getLogger(__name__)

__all__ = [
    "MissingFeatureError",
    "GeometryIsInvalidError",
    "GeometryIsEmptyError",
    "layer_data_generator",
    "feature_to_shapely",
    "parse_q_value",
    "extract_layer_attributes",
    "extract_feature_attributes",
    "is_str_value_null_like",
    "REAL_NONE_JSON_VALUE",
    "NULL_VALUE",
    "NAN_VALUE",
    "STR_NA_VALUE",
    "STR_NONE_VALUE",
    "parse_field",
    "extract_field_value",
    "qgs_geometry_to_shapely",
    "extract_layer_data_single",
]


REAL_NONE_JSON_VALUE = "REAL_NONE"
NAN_VALUE = "nan"
NULL_VALUE = "NULL"
STR_NA_VALUE = "<NA>"
STR_NONE_VALUE = "None"


class MissingFeatureError(Exception): ...


class GeometryIsEmptyError(Exception): ...


class GeometryIsInvalidError(Exception): ...


def parse_q_value(v: Any) -> Any:
    """

    :param v:
    :return:
    """

    from qgis.PyQt.QtCore import QVariant

    from qgis.PyQt.QtGui import QColor

    if isinstance(v, QColor):
        v = v.name()

    elif isinstance(v, QVariant):
        if v.isNull():
            v = None
        else:
            v = v.value()

    return v


def extract_feature_attributes(layer_feature: Any) -> dict[str, Any]:
    """

    :param layer_feature:
    :return:
    """

    return {
        k.name(): parse_q_value(v)
        for k, v in zip(
            layer_feature.fields(),
            layer_feature.attributes(),
        )
    }


def extract_layer_attributes(layer_tree_layer: Any) -> list[dict[str, Any]]:
    """

    :param layer_tree_layer:
    :return:
    """
    geometry_layer = layer_tree_layer.layer()
    if (
        geometry_layer
        and geometry_layer.hasFeatures()
        and geometry_layer.featureCount() > 0
    ):
        layer_feature_attributes = []
        for layer_feature in geometry_layer.getFeatures():
            layer_feature_attributes.append(extract_feature_attributes(layer_feature))

        return layer_feature_attributes

    raise MissingFeatureError(f"no feature was not found for {layer_tree_layer.name()}")


def layer_data_generator(
    layer_tree_layer: Any,
) -> Generator[tuple[dict[str, Any], Any], Any, None]:
    """

    :param layer_tree_layer:
    :return:
    """
    geometry_layer = layer_tree_layer.layer()
    if (
        geometry_layer
        and geometry_layer.hasFeatures()
        and geometry_layer.featureCount() > 0
    ):
        for layer_feature in geometry_layer.getFeatures():
            layer_feature_attributes = extract_feature_attributes(layer_feature)

            if len(layer_feature_attributes) == 0:
                _logger.error(
                    f"Did not find attributes, skipping {layer_tree_layer.name()} {list(geometry_layer.getFeatures())}"
                )
            else:
                _logger.info(
                    f"found {layer_feature_attributes=} for {layer_tree_layer.name()=}"
                )
            yield layer_feature_attributes, layer_feature
    else:
        raise MissingFeatureError(
            f"no feature was not found for {layer_tree_layer.name()}"
        )


def qgs_geometry_to_shapely(
    geom: Any,
    *,
    geom_id: Optional[str] = None,
    validate: bool = True,
) -> Optional[shapely.geometry.base.BaseGeometry]:
    """

    :param geom:
    :param geom_id:
    :param validate:
    :return:
    """
    if geom is not None:
        if validate:
            if not geom.isGeosValid():
                msg = (
                    f"{geom_id} is not a valid geometry, {geom.lastError()}\n"
                    f"{geom.validateGeometry()}"
                )
                _logger.error(msg)

                if True:
                    raise GeometryIsInvalidError(msg)
                else:
                    geom = geom.makeValid()

        if validate:
            if geom.isNull() or geom.isEmpty():
                raise GeometryIsEmptyError(f"{geom_id} is empty")
        else:
            if geom.isNull() or geom.isEmpty():
                return None

        geom_wkb = geom.asWkb()

        if geom_wkb is not None:
            if not isinstance(geom_wkb, bytes):
                geom_wkb = bytes(geom_wkb)

            return shapely.from_wkb(geom_wkb)

    return None


def feature_to_shapely(
    layer_feature: Any,
    validate: bool = True,
) -> Optional[shapely.geometry.base.BaseGeometry]:
    """

    :param validate:
    :param layer_feature:
    :return:
    """
    return qgs_geometry_to_shapely(
        layer_feature.geometry(), geom_id=f"{layer_feature.id()=}", validate=validate
    )


def extract_layer_data_single(layer_tree_layer: Any) -> Tuple:
    """

    :param layer_tree_layer:
    :return:
    """
    geometry_layer = layer_tree_layer.layer()
    if (
        geometry_layer
        and geometry_layer.hasFeatures()
        and geometry_layer.featureCount() > 0
    ):
        if geometry_layer.featureCount() > 1:
            raise ValueError(f"{layer_tree_layer.name()} has more than one feature")

        for layer_feature in geometry_layer.getFeatures():
            layer_feature_attributes = {
                k.name(): parse_q_value(v)
                for k, v in zip(
                    layer_feature.fields(),
                    layer_feature.attributes(),
                )
            }

            if len(layer_feature_attributes) == 0:
                _logger.error(
                    f"Did not find attributes, skipping {layer_tree_layer.name()} {list(geometry_layer.getFeatures())}"
                )
            else:
                _logger.info(
                    f"found {layer_feature_attributes=} for {layer_tree_layer.name()=}"
                )

            return layer_feature_attributes, layer_feature

    raise MissingFeatureError(f"no feature was not found for {layer_tree_layer.name()}")


def parse_field(feature_attributes: Mapping[str, Any], field_name: str) -> Any:
    """

    :param feature_attributes:
    :param field_name:
    :return:
    """
    field_value = feature_attributes[field_name]

    if isinstance(field_value, str):
        ...

    elif isinstance(field_value, QVariant):
        # logger.warning(f"{typeToDisplayString(type(v))}")
        if field_value.isNull():
            field_value = None
        else:
            field_value = field_value.value()

    return field_value


def is_str_value_null_like(v_str_) -> bool:
    """

    :param v_str_:
    :return:
    """
    return (
        (v_str_ == NAN_VALUE.lower())
        or (v_str_ == NULL_VALUE.lower())
        or (v_str_ == STR_NA_VALUE.lower())
        or (v_str_ == STR_NONE_VALUE.lower())
        or len(v_str_.strip()) == 0
    )


def extract_field_value(feature_attributes: Mapping[str, Any], field_name: str) -> Any:
    """

    :param feature_attributes:
    :param field_name:
    :return:
    """
    field_value = feature_attributes.get(field_name)

    if field_value is None:
        ...

    elif isinstance(field_value, QDateTime):
        field_value = field_value.toPyDateTime()

    elif isinstance(field_value, str):
        v = field_value
        v_str = v.lower().strip()

        if is_str_value_null_like(v_str):
            field_value = None

        else:
            field_value = v

    elif isinstance(field_value, QVariant):
        if field_value.isNull():
            field_value = None

        else:
            v = str(field_value.value())

            v_str = v.lower().strip()

            if is_str_value_null_like(v_str):
                field_value = None

            else:
                field_value = v

    return field_value



from qgis.PyQt import QtGui, QtWidgets, uic
from qgis.PyQt.QtCore import Qt
from qgis.PyQt.QtWidgets import QDialog, QHBoxLayout, QLabel, QLineEdit, QWidget

__all__ = [
    "yes_button",
    "no_button",
    "help_button",
    "accept_role",
    "reject_role",
    "echo_mode",
    "horizontal_orientation",
    "window_stays_on_top_hint",
    "item_is_enabled",
    "item_is_selectable",
    "item_is_user_checkable",
    "checked_state",
    "unchecked_state",
    "partially_checked_state",
    "align_center",
    "align_left",
    "align_right",
    "vertical_orientation",
    "message_box_warning",
    "normal_font_weight",
    "normal_font_style",
]

__doc__ = (
    """Provides compatibility between different PyQt versions (often PyQt5 vs PyQt6)"""
)

try:
    yes_button = QtWidgets.QMessageBox.Yes
    no_button = QtWidgets.QMessageBox.No
    help_button = QtWidgets.QMessageBox.Help

except AttributeError:
    yes_button = QtWidgets.QMessageBox.StandardButton.Yes
    no_button = QtWidgets.QMessageBox.StandardButton.No
    help_button = QtWidgets.QMessageBox.StandardButton.Help


try:
    message_box_warning = QtWidgets.QMessageBox.Warning
except AttributeError:
    message_box_warning = QtWidgets.QMessageBox.Icon.Warning


try:
    accept_role = QtWidgets.QMessageBox.ButtonRole.AcceptRole
    reject_role = QtWidgets.QMessageBox.ButtonRole.RejectRole
except AttributeError:
    accept_role = QtWidgets.QMessageBox.AcceptRole
    reject_role = QtWidgets.QMessageBox.RejectRole

try:
    echo_mode = QLineEdit.EchoMode.Password
except AttributeError:
    echo_mode = QLineEdit.Password

try:
    horizontal_orientation = Qt.Orientation.Horizontal
    vertical_orientation = Qt.Orientation.Vertical

except AttributeError:
    horizontal_orientation = Qt.Horizontal
    vertical_orientation = Qt.Vertical

try:
    window_stays_on_top_hint = Qt.WindowType.WindowStaysOnTopHint
except AttributeError:
    window_stays_on_top_hint = Qt.WindowStaysOnTopHint

try:
    checked_state = Qt.CheckState.Checked
    unchecked_state = Qt.CheckState.Unchecked
    partially_checked_state = Qt.CheckState.PartiallyChecked
except AttributeError:
    checked_state = Qt.Checked
    unchecked_state = Qt.Unchecked
    partially_checked_state = Qt.PartiallyChecked

try:
    item_is_enabled = Qt.ItemFlag.ItemIsEnabled
    item_is_selectable = Qt.ItemFlag.ItemIsSelectable
    item_is_user_checkable = Qt.ItemFlag.ItemIsUserCheckable
except AttributeError:
    item_is_enabled = Qt.ItemIsEnabled
    item_is_selectable = Qt.ItemIsSelectable
    item_is_user_checkable = Qt.ItemIsUserCheckable

try:
    align_center = Qt.AlignmentFlag.AlignCenter
    align_left = Qt.AlignmentFlag.AlignLeft
    align_right = Qt.AlignmentFlag.AlignRight
except AttributeError:
    align_center = Qt.AlignCenter
    align_left = Qt.AlignLeft
    align_right = Qt.AlignRight


try:
    normal_font_weight = QtGui.QFont.Normal
    normal_font_style = QtGui.QFont.StyleNormal
except AttributeError:
    normal_font_weight = QtGui.QFont.Weight.Normal
    normal_font_style = QtGui.QFont.Style.StyleNormal
